/// <reference types="node" />
import { EventEmitter } from 'events';
/**
 * Configuration options for the pseudoterminal.
 */
export declare type Options = Partial<ResolvedOptions>;
export interface ResolvedOptions {
    /**
     * Number of columns in the terminal. Defaults to 80.
     */
    cols: number;
    /**
     * Number of rows in the terminal. Defaults to 30.
     */
    rows: number;
    /**
     * Directory where the terminal should start.
     */
    cwd?: string;
    /**
     * Startup shell to use in the terminal. Defaults to the SHELL environment
     * variable in the terminal process.
     */
    shell?: string;
    /**
     * The set of environment variables to set in the terminal process. Defaults
     * to the current process's environment.
     */
    env: {
        [key: string]: string;
    };
}
/**
 * Pseudoterminal control interface
 */
export interface IPty extends EventEmitter {
    /**
     * Number of columns in the terminal.
     */
    cols: number;
    /**
     * Number of rows in the terminal.
     */
    rows: number;
    /**
     * The current working directory of the terminal.
     */
    readonly cwd: string;
    /**
     * The name of the current process running in the terminal.
     */
    readonly process: string;
    /**
     * Write the provided data to the pseudoterminal.
     * @param data Data to write to the terminal.
     */
    write(data: string): void;
    /**
     * Resizes the pseudoterminal to the given dimensions.
     * @param cols Number of columns in the resized terminal.
     * @param rows Number of rows in the resized terminal.
     */
    resize(cols: number, rows: number): void;
    /**
     * Terminate the current pseudoterminal instance.
     */
    kill(): void;
    on(event: string, listener: Function): this;
    on(event: 'data', listener: (data: string) => void): this;
    on(event: 'exit', listener: () => void): this;
    on(event: 'error', listener: (err: any) => void): this;
    emit(event: string | symbol, ...args: any[]): boolean;
    emit(event: 'data', data: string): boolean;
    emit(event: 'exit'): boolean;
    emit(event: 'error', err: any): boolean;
}
/**
 * Create a new pseudoterminal for communicating with the Windows Subsystem for
 * Linux, spawned in a new process.
 * @param options Options to configure the pseudoterminal.
 */
export declare function spawn(options?: Options): IPty;
//# sourceMappingURL=index.d.ts.map