"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const net = require("net");
const path = require("path");
const child_process = require("child_process");
const events_1 = require("events");
const string_decoder_1 = require("string_decoder");
const getPort = require("get-port");
const frame = require("./frame");
class Pty extends events_1.EventEmitter {
    constructor(options) {
        super();
        this._pendingFrames = [];
        this._closed = false;
        this.cols = options.cols;
        this.rows = options.rows;
        this._cwd = options.cwd;
        this._server = net.createServer(socket => {
            // Don't allow more than one socket
            if (this._socket) {
                socket.destroy();
            }
            this._socket = socket;
            // Don't batch up tcp writes
            this._socket.setNoDelay(true);
            const decoder = new string_decoder_1.StringDecoder('utf8');
            this._socket.on('data', data => {
                // TODO: this results in more data copies than needed
                let buf = this._pendingBuffer
                    ? Buffer.concat([this._pendingBuffer, data])
                    : data;
                let f;
                do {
                    f = frame.decode(buf);
                    if (f) {
                        switch (f.type) {
                            case frame.FrameType.Data:
                                this.emit('data', decoder.write(f.data));
                                break;
                            case frame.FrameType.Size:
                                // The backend can't send size frames
                                break;
                            case frame.FrameType.Name:
                                this._procname = f.name;
                                break;
                            case frame.FrameType.Cwd:
                                this._cwd = f.cwd;
                        }
                        if (buf.length - f.size > 0) {
                            this._pendingBuffer = Buffer.alloc(buf.length - f.size);
                            buf.copy(this._pendingBuffer, 0, f.size);
                            buf = this._pendingBuffer;
                        }
                        else {
                            this._pendingBuffer = undefined;
                        }
                    }
                    else {
                        this._pendingBuffer = buf;
                    }
                } while (f && this._pendingBuffer);
            });
            this._socket.on('close', hadError => this.emit('exit'));
            // Flush any pending frames
            for (const f of this._pendingFrames) {
                this._socket.write(f);
            }
            this._pendingFrames = [];
        });
        this._setup(options).catch(e => {
            if (!this._closed) {
                this.emit('error', e);
            }
        });
    }
    get process() {
        return this._procname || '';
    }
    get cwd() {
        return this._cwd || '';
    }
    write(data) {
        const f = frame.encodeData(data);
        if (this._socket) {
            // TODO: use res
            this._socket.write(f);
        }
        else {
            this._pendingFrames.push(f);
        }
    }
    resize(cols, rows) {
        const f = frame.encodeSize(cols, rows);
        if (this._socket) {
            // TODO: use res
            this._socket.write(f);
        }
        else {
            this._pendingFrames.push(f);
        }
    }
    kill() {
        this._closed = true;
        this._server.close();
        if (this._socket) {
            this._socket.destroy();
        }
        if (this._pty && !this._pty.killed) {
            this._pty.kill();
        }
        this.removeAllListeners();
    }
    async _setup(options) {
        const backendPath = await this._getWslPath(path.join(__dirname, '../../backend/target/release/wslpty'));
        const port = await getPort();
        const escapedBackendPath = backendPath.replace(/([\/\\])app\.asar([\/\\])/, '$1app.asar.unpacked$2');
        const args = [escapedBackendPath, String(port)];
        args.push('--cols', String(options.cols));
        args.push('--rows', String(options.rows));
        if (this.cwd) {
            args.push('--cwd', await this._getWslPath(this.cwd));
        }
        if (this._shell) {
            args.push('--shell', this._shell);
        }
        this._server.listen(port, '127.0.0.1', () => {
            this._pty = child_process.spawn('wsl.exe', args, {
                stdio: 'ignore',
                env: options.env
            });
        });
        this._server.on('error', err => {
            if (!this._closed) {
                this.emit('error', err);
            }
        });
        this._server.on('close', () => {
            if (!this._closed) {
                this.emit('exit');
                if (this._pty && !this._pty.killed) {
                    this._pty.kill();
                }
            }
        });
    }
    // TODO: handle nonexistent paths (currently crashes backend)
    async _getWslPath(windowsPath) {
        // Assume linux path if it starts with a slash
        // TODO: beef this up
        if (windowsPath.startsWith('/') || windowsPath.startsWith('~')) {
            return windowsPath;
        }
        return new Promise((resolve, reject) => {
            const proc = child_process.spawn('wsl.exe', 
            // we have to double encode the backslashes
            ['wslpath', windowsPath.replace(/\\/g, '\\\\')]);
            let output = '';
            proc.stdout.on('data', data => output += data);
            proc.on('error', err => {
                reject(err);
            });
            proc.on('close', code => {
                if (code === 0) {
                    resolve(output.trim());
                }
                else {
                    reject(new Error(`Failed to get backend path with error code: ${code}`));
                }
            });
        });
    }
}
/**
 * Create a new pseudoterminal for communicating with the Windows Subsystem for
 * Linux, spawned in a new process.
 * @param options Options to configure the pseudoterminal.
 */
function spawn(options) {
    return new Pty(Object.assign({ cols: 80, rows: 30, env: process.env }, options));
}
exports.spawn = spawn;
//# sourceMappingURL=index.js.map